from __future__ import print_function
import logging
import time
from collections import OrderedDict
from xml.dom import minidom
from xml.parsers.expat import ExpatError
from elifearticle import article as ea
from elifearticle import utils as eautils
from elifetools import utils as etoolsutils
import ejpcsvparser.utils as utils
import ejpcsvparser.csv_data as data

LOGGER = logging.getLogger('parse')
HDLR = logging.FileHandler('parse.log')
FORMATTER = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
HDLR.setFormatter(FORMATTER)
LOGGER.addHandler(HDLR)
LOGGER.setLevel(logging.INFO)


def instantiate_article(article_id):
    LOGGER.info("in instantiate_article for %s", article_id)
    doi = data.get_doi(article_id)
    if doi is not None:
        # Fallback if doi string is blank, default to eLife concatenated
        if doi.strip() == "":
            doi = utils.get_elife_doi(article_id)
        article = ea.Article(doi, title=None)
        return article
    return None


def set_title(article, article_id):
    LOGGER.info("in set_title")
    title = data.get_title(article_id)
    if title:
        article.title = utils.convert_to_xml_string(title)
        return True
    LOGGER.error("could not set title ")
    return False


def set_abstract(article, article_id):
    LOGGER.info("in set_abstract")
    raw_abstract = data.get_abstract(article_id)
    if raw_abstract:
        abstract = utils.decode_cp1252(raw_abstract)
        article.abstract = utils.convert_to_xml_string(abstract)
        article.manuscript = article_id
        return True
    LOGGER.error("could not set abstract ")
    return False


def set_article_type(article, article_id):
    LOGGER.info("in set_article_type")
    article_type_id = data.get_article_type(article_id)
    article_type_index = utils.article_type_indexes()
    if article_type_id in article_type_index:
        article_type = article_type_index[str(article_type_id)]
        article.article_type = article_type['article_type']
        article.display_channel = article_type['display_channel']
        return True
    return False


def set_license(article, article_id):
    LOGGER.info("in set_license")
    # if no article return False
    if not article:
        return False
    license_id = data.get_license(article_id)
    license_object = ea.License(license_id)
    data_values = utils.license_data(license_id)
    # if no data to populate the license return False
    if not data_values:
        return False
    # set the object attributes from the data if present
    for name in ['license_id', 'license_type', 'copyright', 'href',
                 'name', 'paragraph1', 'paragraph2']:
        eautils.set_attr_if_value(license_object, name, data_values.get(name))
    article.license = license_object
    return True


def add_date_to_article(article, date_type, date_string):
    "add a date to the article object"
    if not article:
        return False
    date_struct = None
    date_parts = []
    if date_string:
        date_parts = date_string.split()

    if date_parts:
        try:
            date_struct = time.strptime(date_parts[0], "%Y-%m-%d")
        except ValueError:
            LOGGER.info("unable to convert date %s given %s for article %s",
                        date_type, date_parts, article.doi)
    else:
        return False

    if date_string and date_struct:
        article_date = ea.ArticleDate(date_type, date_struct)
        article.add_date(article_date)
        LOGGER.info("set date_type %s from %s as %s",
                    date_type, date_string, article_date)
        return True
    return False


def set_dates(article, article_id):
    LOGGER.info("in set_dates")
    if not article:
        return False

    accepted_date = data.get_accepted_date(article_id)
    date_status = add_date_to_article(article, "accepted", accepted_date)
    if date_status is not True:
        return False

    received_date = data.get_received_date(article_id)
    if received_date.strip() == "":
        # Use the alternate date column receipt_date if received_date is blank
        received_date = data.get_receipt_date(article_id)
    date_status = add_date_to_article(article, "received", received_date)
    if date_status is not True:
        return False

    # set the license date to be the same as the accepted date
    if article.get_date('accepted'):
        date_license = ea.ArticleDate("license", article.get_date('accepted').date)
        article.add_date(date_license)
    return True


def set_ethics(article, article_id):
    LOGGER.info("in set_ethics")
    ethics = None
    parse_status = None
    ethic = data.get_ethics(article_id)
    LOGGER.info(ethic)
    if ethic:
        parse_status, ethics = parse_ethics(ethic)
    if ethic and parse_status is not True:
        LOGGER.error("could not set ethics due to parsing error")
        return False
    if ethics:
        for ethics_value in ethics:
            article.add_ethic(ethics_value)
    return True


def set_datasets(article, article_id):
    LOGGER.info("in set_datasets")
    datasets = data.get_datasets(article_id)
    dataset_objects = None
    data_availability = None
    parse_status = None
    LOGGER.info(datasets)
    if datasets:
        parse_status, dataset_objects, data_availability = parse_datasets(datasets)
    if datasets and parse_status is not True:
        LOGGER.error("could not set datasets due to parsing error")
        return False
    if dataset_objects:
        for dataset in dataset_objects:
            article.add_dataset(dataset)
    if data_availability:
        article.data_availability = utils.convert_to_xml_string(data_availability)
    return True


def set_categories(article, article_id):
    LOGGER.info("in set_categories")
    categories = data.get_subjects(article_id)
    if categories:
        for category in categories:
            article.add_article_category(category)
    return True


def set_organsims(article, article_id):
    LOGGER.info("in set_organsims")
    research_organisms = data.get_organisms(article_id)
    if research_organisms:
        for research_organism in research_organisms:
            if research_organism.strip() != "":
                article.add_research_organism(utils.convert_to_xml_string(research_organism))
    return True


def set_keywords(article, article_id):
    LOGGER.info("in set_keywords")
    keywords = data.get_keywords(article_id)
    if keywords:
        for keyword in keywords:
            article.add_author_keyword(keyword)
    return True


def build_author(article_id, author_id, author_type):
    "build an author object with the basic name data"
    first_name = utils.decode_cp1252(data.get_author_first_name(article_id, author_id))
    last_name = utils.decode_cp1252(data.get_author_last_name(article_id, author_id))
    middle_name = utils.decode_cp1252(data.get_author_middle_name(article_id, author_id))
    suffix = utils.decode_cp1252(data.get_author_suffix(article_id, author_id))
    # initials = middle_name_initials(middle_name)
    if middle_name.strip() != "":
        # Middle name add to the first name / given name
        first_name += " " + middle_name
    author = ea.Contributor(author_type, last_name, first_name)
    if suffix and suffix.strip() != "":
        author.suffix = suffix
    return author


def author_affiliation(article_id, author_id):
    "create and set author affiliation details"
    affiliation = ea.Affiliation()

    department = utils.decode_cp1252(data.get_author_department(article_id, author_id))
    if department.strip() != "":
        affiliation.department = department
    affiliation.institution = utils.decode_cp1252(
        data.get_author_institution(article_id, author_id))
    city = utils.decode_cp1252(data.get_author_city(article_id, author_id))
    if city.strip() != "":
        affiliation.city = city
    affiliation.country = data.get_author_country(article_id, author_id)

    contrib_type = data.get_author_contrib_type(article_id, author_id)
    dual_corresponding = data.get_author_dual_corresponding(article_id, author_id)
    if (contrib_type == "Corresponding Author" or
            (dual_corresponding.strip() != '' and int(dual_corresponding.strip()) == 1)):
        affiliation.email = data.get_author_email(article_id, author_id)
    return affiliation


def set_author_info(article, article_id):
    """
    author information
    Save the contributor and their position in the list in a dict,
    for both authors and group authors,
    Then add the contributors to the article object in order of their position
    """
    LOGGER.info("in set_author_info")
    authors_dict = {}

    # check there are any authors before continuing
    author_ids = data.get_author_ids(article_id)
    if not author_ids and not data.get_group_authors(article_id):
        LOGGER.error("could not find any author data")
        return False

    if author_ids:
        for author_id in author_ids:

            author_type = "author"
            author = build_author(article_id, author_id, author_type)

            affiliation = author_affiliation(article_id, author_id)
            # set corresponding if the affiliation has an email
            if affiliation.email:
                author.corresp = True

            conflict = data.get_author_conflict(article_id, author_id)
            if conflict.strip() != "":
                author.set_conflict(utils.convert_to_xml_string(conflict))

            orcid = data.get_author_orcid(article_id, author_id)
            if orcid.strip() != "":
                author.orcid = orcid

            author.auth_id = author_id
            author.set_affiliation(affiliation)

            author_position = data.get_author_position(article_id, author_id)
            # Add the author to the dictionary recording their position in the list
            authors_dict[int(author_position)] = author

    # Add group author collab contributors, if present
    group_authors = data.get_group_authors(article_id)
    if group_authors:
        # Parse the group authors string
        group_author_dict = parse_group_authors(group_authors)

        if group_author_dict:
            for author_position in sorted(group_author_dict.keys()):
                collab = group_author_dict.get(author_position)
                author = ea.Contributor("author", None, None, collab)

                # Add the author to the dictionary recording their position in the list
                authors_dict[int(author_position)] = author

    # Finally add authors to the article sorted by their position
    for author_position in sorted(authors_dict.keys()):
        # print article_id, author_position, author
        article.add_contributor(authors_dict.get(author_position))

    return True


def set_editor_info(article, article_id):
    LOGGER.info("in set_editor_info")

    author_type = "editor"

    first_name = utils.decode_cp1252(data.get_me_first_nm(article_id))
    last_name = utils.decode_cp1252(data.get_me_last_nm(article_id))
    middle_name = utils.decode_cp1252(data.get_me_middle_nm(article_id))
    suffix = utils.decode_cp1252(data.get_me_suffix(article_id))
    # no first and last name then return False
    if not(first_name and last_name):
        LOGGER.error("could not set editor")
        return False
    # initials = middle_name_initials(middle_name)
    if middle_name.strip() != "":
        # Middle name add to the first name / given name
        first_name += " " + middle_name
    # create an instance of the POSContributor class
    editor = ea.Contributor(author_type, last_name, first_name)
    if suffix and suffix.strip() != "":
        editor.suffix = suffix
    LOGGER.info("editor is: %s", str(editor))
    LOGGER.info("getting ed id for article %s", article_id)
    LOGGER.info("editor id is %s", data.get_me_id(article_id))
    LOGGER.info(str(type(data.get_me_id(article_id))))
    editor.auth_id = data.get_me_id(article_id)
    affiliation = ea.Affiliation()
    department = data.get_me_department(article_id)
    if department.strip() != "":
        affiliation.department = department
    affiliation.institution = data.get_me_institution(article_id)
    affiliation.country = data.get_me_country(article_id)

    # editor.auth_id = `int(author_id)`we have a me_id, but I need to determine
    # whether that Id is the same as the relevent author id
    editor.set_affiliation(affiliation)
    article.add_contributor(editor)
    return True


def set_funding(article, article_id):
    """
    Instantiate one eLifeFundingAward for each funding award
    Add principal award recipients in the order of author position for the article
    Finally add the funding objects to the article in the order of funding position
    """
    LOGGER.info("in set_funding")
    if not article:
        return False

    # Set the funding note from the manuscript level
    article.funding_note = data.get_funding_note(article_id)

    # Query for all funding award data keys
    funder_ids = data.get_funding_ids(article_id)

    # Keep track of funding awards by position in a dict
    funding_awards = OrderedDict()

    # First pass, build the funding awards
    if funder_ids:
        for (funder_article_id, author_id, funder_position) in funder_ids:
            # print (article_id, author_id, funder_position)
            funder_identifier = data.get_funder_identifier(
                funder_article_id, author_id, funder_position)
            funder = utils.decode_cp1252(utils.clean_funder(
                data.get_funder(funder_article_id, author_id, funder_position)))
            award_id = data.get_award_id(funder_article_id, author_id, funder_position)

            if funder_position not in funding_awards.keys():
                # Initialise the object values
                funding_awards[funder_position] = ea.FundingAward()
                if funder:
                    funding_awards[funder_position].institution_name = funder
                if funder_identifier and funder_identifier.strip() != "":
                    funding_awards[funder_position].institution_id = funder_identifier
                if award_id and award_id.strip() != "":
                    funding_awards[funder_position].add_award_id(award_id)

    # Second pass, add the primary award recipients in article author order
    for position in sorted(funding_awards.keys()):
        for contrib in article.contributors:
            for (funder_article_id, author_id, funder_position) in funder_ids:
                if position == funder_position and contrib.auth_id == author_id:
                    funding_awards[position].add_principal_award_recipient(contrib)

    # Add funding awards to the article object, sorted by position
    for position in sorted(funding_awards.keys()):
        article.add_funding_award(funding_awards.get(position))
    return True


def parse_ethics(ethic):
    """
    Given angle bracket escaped XML string, parse
    animal and human ethic comments, and return
    a list of strings if involved_comments tag
    is found. Boiler plate prefix added too.
    """

    ethics = []
    reparsed = None
    parse_status = None

    # Decode escaped angle brackets
    LOGGER.info("ethic is %s", ethic)
    ethic_xml = utils.unserialise_angle_brackets(ethic)
    ethic_xml = etoolsutils.escape_ampersand(ethic_xml)
    LOGGER.info("ethic is %s", ethic_xml)

    # Parse XML
    try:
        reparsed = minidom.parseString(ethic_xml)
        parse_status = True
    except ExpatError:
        parse_status = False
        LOGGER.info("ethic reparsed is %s", reparsed)

    # Extract comments
    if reparsed:
        for ethic_type in 'animal_subjects', 'human_subjects':
            ethic_node = reparsed.getElementsByTagName(ethic_type)[0]
            for node in ethic_node.childNodes:
                if node.nodeName == 'involved_comments':
                    text_node = node.childNodes[0]
                    ethic_text = text_node.nodeValue

                    # Add boilerplate
                    if ethic_type == 'animal_subjects':
                        ethic_text = 'Animal experimentation: ' + ethic_text.strip()
                    elif ethic_type == 'human_subjects':
                        ethic_text = 'Human subjects: ' + ethic_text.strip()

                    # Decode unicode characters
                    ethics.append(utils.entity_to_unicode(ethic_text))

    return parse_status, ethics


def parse_dataset_node(dataset_node, dataset_type):
    "extract attributes from a minidom node and populate a Dataset object"
    dataset = ea.Dataset()

    dataset.dataset_type = dataset_type

    for node in dataset_node.childNodes:

        if node.nodeName == 'authors_text_list' and node.childNodes:
            text_node = node.childNodes[0]
            for author_name in text_node.nodeValue.split(','):
                if author_name.strip() != '':
                    dataset.add_author(author_name.lstrip())

        if node.nodeName == 'title':
            text_node = node.childNodes[0]
            dataset.title = utils.entity_to_unicode(text_node.nodeValue)

        if node.nodeName == 'id':
            text_node = node.childNodes[0]
            dataset.source_id = utils.entity_to_unicode(text_node.nodeValue)

        if node.nodeName == 'license_info':
            text_node = node.childNodes[0]
            dataset.license_info = utils.entity_to_unicode(text_node.nodeValue)

        if node.nodeName == 'year' and node.childNodes:
            text_node = node.childNodes[0]
            dataset.year = utils.entity_to_unicode(text_node.nodeValue)

    return dataset


def parse_datasets(datasets_content):
    """
    Datasets content is XML with escaped angle brackets
    """
    datasets = []
    data_availability = None
    reparsed = None
    parse_status = None

    # Decode escaped angle brackets
    LOGGER.info("datasets is %s", datasets_content)
    datasets_xml = utils.escape_angle_brackets(datasets_content)
    datasets_xml = utils.unserialise_angle_brackets(datasets_xml)
    datasets_xml = etoolsutils.escape_ampersand(datasets_xml)
    LOGGER.info("datasets is %s", datasets_xml)

    # Parse XML
    try:
        reparsed = minidom.parseString(datasets_xml)
        parse_status = True
    except ExpatError:
        LOGGER.info("datasets reparsed is %s", reparsed)
        parse_status = False

    # Extract comments
    if reparsed:
        for dataset_type in 'datasets', 'prev_published_datasets':
            datasets_nodes = reparsed.getElementsByTagName(dataset_type)[0]
            for dataset_node in datasets_nodes.getElementsByTagName("dataset"):
                datasets.append(parse_dataset_node(dataset_node, dataset_type))

        # Parse the data availability statement
        if reparsed.getElementsByTagName('data_availability_textbox'):
            data_availability_node = reparsed.getElementsByTagName('data_availability_textbox')
            if data_availability_node[0].childNodes:
                data_availability = utils.entity_to_unicode(
                    data_availability_node[0].childNodes[0].nodeValue)

    return parse_status, datasets, data_availability


def parse_group_authors(group_authors):
    """
    Given a raw group author value from the data files,
    check for empty, whitespace, zero
    If not empty, remove extra numbers from the end of the string
    Return a dictionary of dict[author_position] = collab_name
    """
    group_author_dict = OrderedDict()
    if not group_authors:
        group_author_dict = None
    elif group_authors.strip() == "" or group_authors.strip() == "0":
        group_author_dict = None
    else:

        # Parse out elements into a list, clean and
        #  add the the dictionary using some steps

        # Split the string on the first delimiter
        group_author_list = group_authors.split('order_start')

        for group_author_string in group_author_list:
            if group_author_string == "":
                continue

            # Now split on the second delimiter
            position_and_name = group_author_string.split('order_end')

            author_position = position_and_name[0]

            # Strip numbers at the end
            if len(position_and_name) > 1:
                group_author = position_and_name[1].rstrip("1234567890")

                # Finally, add to the dict noting the authors position
                group_author_dict[author_position] = group_author

    return group_author_dict


def build_article(article_id):
    """
    Given an article_id, instantiate and populate the article objects
    """
    error_count = 0
    error_messages = []

    # Only happy with string article_id - cast it now to be safe!
    article_id = str(article_id)

    article = instantiate_article(article_id)

    # Run each of the below functions to build the article object components
    article_set_functions = [
        set_title, set_abstract, set_article_type, set_license, set_dates, set_ethics,
        set_datasets, set_categories, set_organsims, set_author_info, set_editor_info,
        set_keywords, set_funding]
    for set_function in article_set_functions:
        if not set_function(article, article_id):
            error_count = error_count + 1
            error_messages.append("article_id " + str(article_id)
                                  + " error in " + set_function.__name__)

    # Building from CSV data it must be a POA type, set it
    if article:
        article.is_poa = True

    print(error_count)

    # default conflict text
    if article:
        article.conflict_default = "The authors declare that no competing interests exist."

    if error_count == 0:
        return article, error_count, error_messages

    return None, error_count, error_messages



import re
from collections import OrderedDict
from elifetools import utils as etoolsutils
from elifearticle import utils as eautils
import ejpcsvparser.settings as settings


def allowed_tags():
    "tuple of whitelisted tags"
    return (
        '<i>', '</i>',
        '<italic>', '</italic>',
        '<b>', '</b>',
        '<bold>', '</bold>',
        '<sup>', '</sup>',
        '<sub>', '</sub>',
        '<u>', '</u>',
        '<underline>', '</underline>',
        '<b>', '</b>',
        '<bold>', '</bold>',
        '<p>', '</p>',
    )


def article_type_indexes():
    "boilerplate article-type values based on id in CSV file"
    article_type_index = OrderedDict()
    article_type_index['1'] = {
        'article_type':    'research-article',
        'display_channel': 'Research Article'}
    article_type_index['8'] = {
        'article_type':    'discussion',
        'display_channel': 'Feature Article'}
    article_type_index['10'] = {
        'article_type':    'research-article',
        'display_channel': 'Feature Article'}
    article_type_index['14'] = {
        'article_type':    'research-article',
        'display_channel': 'Short Report'}
    article_type_index['15'] = {
        'article_type':    'research-article',
        'display_channel': 'Research Advance'}
    article_type_index['19'] = {
        'article_type':    'research-article',
        'display_channel': 'Tools and Resources'}
    article_type_index['21'] = {
        'article_type':    'research-article',
        'display_channel': 'Scientific Correspondence'}
    return article_type_index


def license_data(license_id):
    "boilerplate data to populate a license object keyed on the license_id"
    license_data = OrderedDict()
    if not license_id:
        return license_data
    if int(license_id) == 1:
        license_data['license_id'] = license_id
        license_data['license_type'] = "open-access"
        license_data['copyright'] = True
        license_data['href'] = "http://creativecommons.org/licenses/by/4.0/"
        license_data['name'] = "Creative Commons Attribution License"
        license_data['paragraph1'] = "This article is distributed under the terms of the "
        license_data['paragraph2'] = (
            " permitting unrestricted use and redistribution provided that the " +
            "original author and source are credited.")
    elif int(license_id) == 2:
        license_data['license_id'] = license_id
        license_data['license_type'] = "open-access"
        license_data['copyright'] = False
        license_data['href'] = "http://creativecommons.org/publicdomain/zero/1.0/"
        license_data['name'] = "Creative Commons CC0"
        license_data['paragraph1'] = (
            "This is an open-access article, free of all copyright, and may be " +
            "freely reproduced, distributed, transmitted, modified, built upon, or " +
            "otherwise used by anyone for any lawful purpose. The work is made " +
            "available under the ")
        license_data['paragraph2'] = " public domain dedication."
    return license_data


def entity_to_unicode(string):
    """
    Quick convert unicode HTML entities to unicode characters
    using a regular expression replacement
    """
    if not string:
        return string
    # Selected character replacements that have been seen
    replacements = []
    replacements.append((r"&alpha;", u"\u03b1"))
    replacements.append((r"&beta;", u"\u03b2"))
    replacements.append((r"&gamma;", u"\u03b3"))
    replacements.append((r"&delta;", u"\u03b4"))
    replacements.append((r"&epsilon;", u"\u03b5"))
    replacements.append((r"&ordm;", u"\u00ba"))
    replacements.append((r"&iuml;", u"\u00cf"))
    replacements.append((r"&ldquo;", '"'))
    replacements.append((r"&rdquo;", '"'))

    # First, replace numeric entities with unicode
    string = re.sub(r"&#x(....);", eautils.repl, string)
    # Second, replace some specific entities specified in the list
    for entity, replacement in replacements:
        string = re.sub(entity, replacement, string)
    return string


def entities(function):
    """
    Convert entities to unicode as a decorator
    """
    def wrapper(*args, **kwargs):
        value = function(*args, **kwargs)
        return entity_to_unicode(value)
    return wrapper


def decode_brackets(string):
    """
    Decode angle bracket escape sequence
    used to encode XML content
    """
    string = str(string)
    string = string.replace(settings.LESS_THAN_ESCAPE_SEQUENCE, '<')
    string = string.replace(settings.GREATER_THAN_ESCAPE_SEQUENCE, '>')
    return string


def unserialise_angle_brackets(escaped_string):
    unserial_xml = escaped_string.replace(settings.LESS_THAN_ESCAPE_SEQUENCE, "<")
    unserial_xml = unserial_xml.replace(settings.GREATER_THAN_ESCAPE_SEQUENCE, ">")
    return unserial_xml


def convert_to_xml_string(string):
    """
    For input strings with escaped tags and special characters
    issue a set of conversion functions to prepare it prior
    to adding it to an article object
    """
    string = entity_to_unicode(string)
    string = decode_brackets(string)
    string = eautils.replace_tags(string, 'i', 'italic')
    string = eautils.replace_tags(string, 'u', 'underline')
    string = eautils.replace_tags(string, 'b', 'bold')
    string = eautils.replace_tags(string, 'em', 'italic')
    string = etoolsutils.escape_unmatched_angle_brackets(string, allowed_tags())
    return string


def escape_angle_brackets(string):
    "replace angle brackets only for when parsing escaped XML from CSV files"
    try:
        return string.replace('<', '&lt;').replace('>', '&gt;')
    except AttributeError:
        return string


def get_elife_doi(article_id):
    """
    Given an article_id, return a DOI for the eLife journal
    """
    doi = "10.7554/eLife." + str(int(article_id)).zfill(5)
    return doi


def decode_cp1252(string):
    """
    CSV files look to be in CP-1252 encoding (Western Europe)
    Decoding to ASCII is normally fine, except when it gets an O umlaut, for example
    In this case, values must be decoded from cp1252 in order to be added as unicode
    to the final XML output.
    This function helps do that in selected places, like on author surnames
    """
    if not string:
        return string
    try:
        # See if it is not safe to encode to ascii first
        string.encode('ascii')
    except (UnicodeEncodeError, UnicodeDecodeError):
        # Wrap the decode in another exception to make sure this never fails
        try:
            string = string.decode('cp1252')
        except (UnicodeEncodeError, UnicodeDecodeError, AttributeError):
            pass
    return string


def clean_funder(funder):
    """
    Remove extra content from funder names
    separated by | character
    and anything in parentheses
    """
    funder = funder.split('|')[-1]
    funder = re.sub(r"\(.*\)", "", funder)
    funder = funder.rstrip().lstrip()
    return funder



import logging
import csv
import sys
import io
import os
from collections import defaultdict, OrderedDict
import ejpcsvparser.utils as utils
import ejpcsvparser.settings as settings

# todo!! clean up these values and the settings
CSV_PATH = settings.CSV_PATH
TMP_DIR = settings.TMP_DIR
ROWS_WITH_COLNAMES = settings.ROWS_WITH_COLNAMES
DATA_START_ROW = settings.DATA_START_ROW
CSV_FILES = settings.CSV_FILES
COLUMN_HEADINGS = settings.CSV_COLUMN_HEADINGS
OVERFLOW_CSV_FILES = settings.OVERFLOW_CSV_FILES

LOGGER = logging.getLogger('csv_data')
HDLR = logging.FileHandler('csv_data.log')
FORMATTER = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
HDLR.setFormatter(FORMATTER)
LOGGER.addHandler(HDLR)
LOGGER.setLevel(logging.INFO)


def memoize(value):
    "Memoization decorator for functions taking one or more arguments."
    class Memodict(dict):
        "Memoization dict"
        def __init__(self, value):
            dict.__init__(self)
            self.value = value

        def __call__(self, *args):
            return self[args]

        def __missing__(self, key):
            ret = self[key] = self.value(*key)
            return ret
    return Memodict(value)


def get_csv_path(path_type):
    """
    sets the location of the path to the author csv file
    returns the path

    This is the only function where the path the our actual data files
    are set.
    """
    path = CSV_PATH + CSV_FILES[path_type]
    return path


@memoize
def get_csv_col_names(table_type):
    LOGGER.info("in get_csv_col_names")
    LOGGER.info(table_type)
    sheet = get_csv_sheet(table_type)
    LOGGER.info(sheet)
    LOGGER.info(str(ROWS_WITH_COLNAMES))
    columns_row = []
    for index, row in enumerate(sheet):
        LOGGER.info("in enumerate")
        LOGGER.info(str(index) + " " + str(row))
        LOGGER.debug(str(index) + " " + str(ROWS_WITH_COLNAMES))
        if int(index) == int(ROWS_WITH_COLNAMES):
            columns_row = row
    return columns_row


@memoize
def get_csv_data_rows(table_type):
    sheet = get_csv_sheet(table_type)
    rows = []
    for row in sheet:
        rows.append(row)
    data_rows = rows[DATA_START_ROW:]
    return data_rows


def get_cell_value(col_name, col_names, row):
    """
    we pass the name of the col and a copy of the col names row in to
    this fucntion so that we don't have to worry about knowing what the
    index of a specific col name is.
    """
    position = col_names.index(col_name)
    if row and position and len(row) > position:
        return row[position]
    return None


def join_lines(line_one, line_two, line_number, data_start_row=DATA_START_ROW):
    "join multiple lines together taking into account the header rows"
    if line_number <= data_start_row:
        # keep blank lines found in the headers
        content = line_two
    else:
        if line_two.lstrip() == '':
            # blank line outside of the header convert to a space
            content = line_one.rstrip("\r\n") + u' '
        else:
            content = line_one.rstrip("\r\n") + line_two.lstrip()
    return content


def do_add_line(content, line_number, data_start_row=DATA_START_ROW):
    "decide if the line should be added to the output"
    add_line = False
    if line_number <= data_start_row or content.rstrip().endswith('"'):
        add_line = True
    return add_line


def flatten_lines(iterable, data_start_row=DATA_START_ROW):
    "iterate through an open file and join lines"
    clean_csv_data = u''
    line_number = 1
    prev_line = u''
    add_line = False
    for content in iterable:
        content = utils.decode_cp1252(content)
        # add the line based on the previous iteration value
        if add_line:
            clean_csv_data += prev_line
            prev_line = u''
        prev_line = join_lines(prev_line, content, line_number, data_start_row)
        add_line = do_add_line(content, line_number, data_start_row)
        line_number += 1
    # Add the final line
    clean_csv_data += prev_line
    return clean_csv_data


@memoize
def clean_csv(path):
    "fix CSV file oddities making it difficult to parse"
    clean_csv_data = u''
    new_path = os.path.join(TMP_DIR, os.path.split(path)[-1])
    with open(path, 'r') as open_read_file:
        clean_csv_data = flatten_lines(open_read_file)
    with open(new_path, 'w') as open_write_file:
        open_write_file.write(clean_csv_data)
    return new_path


@memoize
def get_csv_sheet(table_type):
    LOGGER.info("in get_csv_sheet")
    path = get_csv_path(table_type)
    LOGGER.info(str(path))

    path = clean_csv(path)

    # https://docs.python.org/3/library/functions.html#open
    handle = io.open(path, 'r', newline='', encoding='utf-8', errors='surrogateescape')

    with handle as csvfile:
        csvreader = csv.reader(csvfile, delimiter=',', quotechar='"')
        sheet = []
        for row in csvreader:
            sheet.append(row)
    # For overflow file types, parse again with no quotechar
    if table_type in OVERFLOW_CSV_FILES:
        csvfile = open(path)
        csvreader = csv.reader(csvfile, delimiter=',', quotechar=None)
        if table_type in ["ethics", "datasets"]:
            join_cells_from = 3
        else:
            join_cells_from = 2
        for row in csvreader:
            if csvreader.line_num <= DATA_START_ROW:
                continue
            # Merge cells 3 to the end because any commas will cause extra columns
            row[join_cells_from] = ','.join(row[join_cells_from:])
            for index, cell in enumerate(row):
                # Strip leading quotation marks
                row[index] = cell.lstrip('"').rstrip('"')
            sheet[csvreader.line_num-1] = row
        csvfile.close()
    return sheet


@memoize
def index_table_on_article_id(table_type):
    """
    return a dict of the CSV file keyed on article_id

    the name of the manuscript number column is hard wired in this function.
    """

    LOGGER.info("in index_table_on_article_id")

    # get the data and the row of colnames
    data_rows = get_csv_data_rows(table_type)
    col_names = get_csv_col_names(table_type)

    # LOGGER.info("data_rows: " + str(data_rows))
    LOGGER.info("col_names: %s", col_names)

    article_index = defaultdict(list)
    for data_row in data_rows:
        article_id = get_cell_value('poa_m_ms_no', col_names, data_row)
        # author_id = get_cell_value("poa_a_id", col_names, data_row)
        article_index[article_id].append(data_row)
        # print article_id, author_id
    return article_index


@memoize
def index_authors_on_article_id():
    article_index = index_table_on_article_id("authors")
    return article_index


@memoize
def index_authors_on_author_id():
    # """
    # as we are going to be doing a lot of looking up authors by
    # author_id and manuscript_id,
    # so we are going to make a dict of dicts indexed on manuscript id and then author id
    # """
    table_type = "authors"
    col_names = get_csv_col_names(table_type)
    author_table = index_authors_on_article_id()

    article_ids = author_table.keys()
    article_author_index = OrderedDict()  # this is the key item we will return our of this function
    for article_id in article_ids:
        rows = author_table[article_id]
        author_index = defaultdict()
        for row in rows:
            author_id = get_cell_value("poa_a_id", col_names, row)
            author_index[author_id] = row
        article_author_index[article_id] = author_index
    return article_author_index


@memoize
def get_article_attributes(article_id, attribute_type, attribute_label):
    LOGGER.info("in get_article_attributes")
    LOGGER.info("article_id: %s attribute_type: %s attribute_label: %s",
                article_id, attribute_type, attribute_label)
    attributes = []
    LOGGER.info("about to generate attribute index")
    attribute_index = index_table_on_article_id(attribute_type)
    LOGGER.info("generated attribute index")
    # LOGGER.info(str(attribute_index))
    LOGGER.info("about to get col_names for colname %s", attribute_type)
    col_names = get_csv_col_names(attribute_type)
    attribute_rows = attribute_index[str(article_id)]
    for attribute_row in attribute_rows:
        attributes.append(get_cell_value(attribute_label, col_names, attribute_row))
    return attributes


def article_all_values(article_id, file_name, column_name):
    "get the article attributes and return all of them"
    return get_article_attributes(article_id, file_name, column_name)


def article_first_value(article_id, file_name, column_name):
    "get the article attributes and return only the first value"
    attributes = article_all_values(article_id, file_name, column_name)
    if attributes:
        return attributes[0]
    return None


# subjects table
def get_subjects(article_id):
    return article_all_values(article_id, "subjects", COLUMN_HEADINGS["subject_areas"])


# organisms table
def get_organisms(article_id):
    return article_all_values(article_id, "organisms", COLUMN_HEADINGS["organisms"])


# license table
def get_license(article_id):
    return article_first_value(article_id, "license", COLUMN_HEADINGS["license_id"])


# keywords table
def get_keywords(article_id):
    return article_all_values(article_id, "keywords", COLUMN_HEADINGS["keywords"])


# manuscript table
@utils.entities
def get_title(article_id):
    return article_first_value(article_id, "title", COLUMN_HEADINGS["title"])


@utils.entities
def get_abstract(article_id):
    return article_first_value(article_id, "abstract", COLUMN_HEADINGS["abstract"])


def get_doi(article_id):
    return article_first_value(article_id, "manuscript", COLUMN_HEADINGS["doi"])


def get_article_type(article_id):
    return article_first_value(article_id, "manuscript", COLUMN_HEADINGS["article_type"])


def get_accepted_date(article_id):
    return article_first_value(article_id, "manuscript", COLUMN_HEADINGS["accepted_date"])


def get_received_date(article_id):
    return article_first_value(article_id, "received", COLUMN_HEADINGS["received_date"])


def get_receipt_date(article_id):
    return article_first_value(article_id, "received", COLUMN_HEADINGS["receipt_date"])


def get_me_id(article_id):
    return article_first_value(article_id, "manuscript", COLUMN_HEADINGS["editor_id"])


@utils.entities
def get_me_last_nm(article_id):
    return article_first_value(article_id, "manuscript", COLUMN_HEADINGS["editor_last_name"])


@utils.entities
def get_me_first_nm(article_id):
    return article_first_value(article_id, "manuscript", COLUMN_HEADINGS["editor_first_name"])


@utils.entities
def get_me_middle_nm(article_id):
    return article_first_value(article_id, "manuscript", COLUMN_HEADINGS["editor_middle_name"])


@utils.entities
def get_me_suffix(article_id):
    try:
        return article_first_value(article_id, "manuscript", COLUMN_HEADINGS["editor_suffix"])
    except ValueError:
        return None


@utils.entities
def get_me_institution(article_id):
    return article_first_value(article_id, "manuscript", COLUMN_HEADINGS["editor_institution"])


@utils.entities
def get_me_department(article_id):
    return article_first_value(article_id, "manuscript", COLUMN_HEADINGS["editor_department"])


@utils.entities
def get_me_country(article_id):
    return article_first_value(article_id, "manuscript", COLUMN_HEADINGS["editor_country"])


def get_ethics(article_id):
    """
    needs a bit of refinement owing to serilaising of data by EJP
    """
    return article_first_value(article_id, "ethics", COLUMN_HEADINGS["ethics"])


# authors table
def get_author_ids(article_id):
    return article_all_values(article_id, "authors", COLUMN_HEADINGS["author_id"])


def get_author_attribute(article_id, author_id, attribute_name):
    article_author_index = index_authors_on_author_id()
    # check for if the data row exists first
    if article_id not in article_author_index:
        return None
    if author_id not in article_author_index[article_id]:
        return None
    # continue
    data_row = article_author_index[article_id][author_id]
    col_names = get_csv_col_names("authors")
    attribute = get_cell_value(attribute_name, col_names, data_row)
    return attribute


def get_author_position(article_id, author_id):
    return get_author_attribute(article_id, author_id, COLUMN_HEADINGS["author_position"])


def get_author_email(article_id, author_id):
    return get_author_attribute(article_id, author_id, COLUMN_HEADINGS["email"])


def get_author_contrib_type(article_id, author_id):
    return get_author_attribute(article_id, author_id, COLUMN_HEADINGS["author_type"])


def get_author_dual_corresponding(article_id, author_id):
    return get_author_attribute(article_id, author_id, COLUMN_HEADINGS["dual_corresponding"])


@utils.entities
def get_author_last_name(article_id, author_id):
    return get_author_attribute(article_id, author_id, COLUMN_HEADINGS["author_last_name"])


@utils.entities
def get_author_first_name(article_id, author_id):
    return get_author_attribute(article_id, author_id, COLUMN_HEADINGS["author_first_name"])


@utils.entities
def get_author_middle_name(article_id, author_id):
    return get_author_attribute(article_id, author_id, COLUMN_HEADINGS["author_middle_name"])


@utils.entities
def get_author_suffix(article_id, author_id):
    try:
        return get_author_attribute(article_id, author_id, COLUMN_HEADINGS["author_suffix"])
    except ValueError:
        return None


@utils.entities
def get_author_institution(article_id, author_id):
    return get_author_attribute(article_id, author_id, COLUMN_HEADINGS["author_institution"])


@utils.entities
def get_author_department(article_id, author_id):
    return get_author_attribute(article_id, author_id, COLUMN_HEADINGS["author_department"])


@utils.entities
def get_author_city(article_id, author_id):
    return get_author_attribute(article_id, author_id, COLUMN_HEADINGS["author_city"])


@utils.entities
def get_author_country(article_id, author_id):
    return get_author_attribute(article_id, author_id, COLUMN_HEADINGS["author_country"])


def get_author_state(article_id, author_id):
    return get_author_attribute(article_id, author_id, COLUMN_HEADINGS["author_state"])


def get_author_conflict(article_id, author_id):
    return get_author_attribute(article_id, author_id, COLUMN_HEADINGS["author_conflict"])


def get_author_orcid(article_id, author_id):
    return get_author_attribute(article_id, author_id, COLUMN_HEADINGS["orcid"])


def get_group_authors(article_id):
    return article_first_value(article_id, "group_authors", COLUMN_HEADINGS["group_author"])


def get_datasets(article_id):
    return article_first_value(article_id, "datasets", COLUMN_HEADINGS["datasets"])


# funding
@memoize
def index_funding_table():
    """
    Rows in the funding CSV are to be uniquely identified by three column values
    article_id + author_id + funder_position
    This will return a three dimensional dict with those hierarchies
    """
    table_type = "funding"

    LOGGER.info("in index_funding_table")

    # get the data and the row of colnames
    data_rows = get_csv_data_rows(table_type)
    col_names = get_csv_col_names(table_type)

    # LOGGER.info("data_rows: " + str(data_rows))
    LOGGER.info("col_names: %s", col_names)

    article_index = OrderedDict()
    for data_row in data_rows:
        article_id = get_cell_value('poa_m_ms_no', col_names, data_row)
        author_id = get_cell_value(COLUMN_HEADINGS["author_id"], col_names, data_row)
        funder_position = get_cell_value(COLUMN_HEADINGS["funder_position"], col_names, data_row)

        # Crude multidimentional dict builder
        if article_id not in article_index:
            article_index[article_id] = OrderedDict()
        if author_id not in article_index[article_id]:
            article_index[article_id][author_id] = OrderedDict()

        article_index[article_id][author_id][funder_position] = data_row

    return article_index


def get_funding_ids(article_id):
    """
    Return funding table keys as a list of tuples
    for a particular article_id
    """
    funding_ids = []

    for key, value in index_funding_table().items():
        if key == article_id:
            for key_2, value_2 in value.items():
                for key_3 in value_2.keys():
                    funding_ids.append((key, key_2, key_3))

    return funding_ids


def get_funding_attribute(article_id, author_id, funder_position, attribute_name):
    funding_article_index = index_funding_table()

    data_row = funding_article_index[str(article_id)][str(author_id)][str(funder_position)]

    col_names = get_csv_col_names("funding")
    attribute = get_cell_value(attribute_name, col_names, data_row)
    return attribute


def get_funder(article_id, author_id, funder_position):
    return get_funding_attribute(
        article_id, author_id, funder_position, COLUMN_HEADINGS["funder"])


def get_award_id(article_id, author_id, funder_position):
    return get_funding_attribute(
        article_id, author_id, funder_position, COLUMN_HEADINGS["award_id"])


def get_funder_identifier(article_id, author_id, funder_position):
    return get_funding_attribute(
        article_id, author_id, funder_position, COLUMN_HEADINGS["funder_identifier"])


def get_funding_note(article_id):
    return article_first_value(article_id, "manuscript", COLUMN_HEADINGS["funding_note"])



from setuptools import setup
import ejpcsvparser

with open("README.md") as fp:
    README = fp.read()

setup(
    name="ejpcsvparser",
    version=ejpcsvparser.__version__,
    description="EJP CSV parser for building article objects.",
    long_description=README,
    long_description_content_type="text/markdown",
    packages=["ejpcsvparser"],
    license="MIT",
    install_requires=["elifearticle", "GitPython", "configparser"],
    url="https://github.com/elifesciences/ejp-csv-parser",
    maintainer="eLife Sciences Publications Ltd.",
    maintainer_email="tech-team@elifesciences.org",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
    ],
)




# csv reading settings
ROWS_WITH_COLNAMES = 3
DATA_START_ROW = 4
LESS_THAN_ESCAPE_SEQUENCE = 'LTLT'
GREATER_THAN_ESCAPE_SEQUENCE = 'GTGT'

CSV_PATH = "tests/test_data/"

TMP_DIR = "tests/tmp/"

CSV_FILES = {
    "authors": "poa_author.csv",
    "license": "poa_license.csv",
    "manuscript": "poa_manuscript.csv",
    "received": "poa_received.csv",
    "subjects": "poa_subject_area.csv",
    "organisms": "poa_research_organism.csv",
    "abstract": "poa_abstract.csv",
    "title": "poa_title.csv",
    "keywords": "poa_keywords.csv",
    "group_authors": "poa_group_authors.csv",
    "datasets": "poa_datasets.csv",
    "funding": "poa_funding.csv",
    "ethics": "poa_ethics.csv"
}

# Special files that allow quotation marks in their final column: column 3
OVERFLOW_CSV_FILES = ["abstract", "title", "ethics", "datasets"]

CSV_COLUMN_HEADINGS = {
    "author_position": "poa_a_seq",
    "subject_areas": "poa_s_subjectarea",
    "license_id": "poa_l_license_id",
    "title": "poa_m_title_tag",
    "abstract": "poa_m_abstract_tag",
    "doi": "poa_m_doi",
    "article_type": "poa_m_type",
    "accepted_date": "poa_m_accepted_dt",
    "received_date": "poa_r_received_dt",
    "receipt_date": "poa_r_receipt_dt2",
    "editor_id": "poa_m_me_id",
    "editor_last_name": "poa_m_me_last_nm",
    "editor_first_name": "poa_m_me_first_nm",
    "editor_middle_name": "poa_m_me_middle_nm",
    "editor_suffix": "poa_m_me_suffix",
    "editor_institution": "poa_m_me_organization",
    "editor_department": "poa_m_me_department",
    "editor_country": "poa_m_me_country",
    "ethics": "poa_m_ethics_note",
    "author_id": "poa_a_id",
    "email": "poa_a_email",
    "author_type": "poa_a_type_cde",
    "dual_corresponding": "poa_a_dual_corr",
    "author_last_name": "poa_a_last_nm",
    "author_first_name": "poa_a_first_nm",
    "author_middle_name": "poa_a_middle_nm",
    "author_suffix": "poa_a_suffix",
    "author_institution": "poa_a_organization",
    "author_department": "poa_a_department",
    "author_city": "poa_a_city",
    "author_country": "poa_a_country",
    "author_state": "poa_a_state",
    "author_conflict": "poa_a_cmp",
    "organisms": "poa_ro_researchorganism",
    "keywords": "poa_kw_keyword",
    "group_author": "poa_ga",
    "orcid": "ORCID",
    "datasets": "poa_m_dataset_note",
    "award_id": "poa_grant_ref_no",
    "funder_position": "poa_funder_order",
    "funder": "poa_funder",
    "funder_identifier": "poa_fund_ref_id",
    "funding_note": "poa_m_funding_note"
}
